package com.yuandian.dataflow.controller;

import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.entity.PeerId;
import com.alipay.sofa.jraft.error.RemotingException;
import com.alipay.sofa.jraft.rpc.InvokeCallback;
import com.google.protobuf.Any;
import com.yuandian.dataflow.controller.PacketsProcessor.PacketsRequest;
import com.yuandian.dataflow.proto.msgtype.BacktrackingFlowOuterClass;
import com.yuandian.dataflow.statemachine.StateFactory;
import com.yuandian.dataflow.statemachine.annotations.MasterRegister;
import com.yuandian.dataflow.statemachine.closure.GenericClosure;
import com.yuandian.dataflow.statemachine.master.MasterContext;
import com.yuandian.dataflow.statemachine.master.MasterExecute;
import com.yuandian.dataflow.statemachine.operate.Operate;
import com.yuandian.dataflow.statemachine.operate.Operate.OperateType;
import com.yuandian.dataflow.statemachine.state.State;
import com.yuandian.dataflow.statemachine.state.WorkerState;
import com.yuandian.dataflow.utils.PacketsManager;
import com.yuandian.dataflow.utils.Utils;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@MasterRegister
public class MasterProcessor implements MasterExecute {

    /**
     * 最大任务数限制
     */
    private final int MAX_TASKS = 1000;
    private final int DEFAULT_ASYNC_TIMEOUT = 5000;

    public static PacketsManager packetsManager = new PacketsManager();

    @Override
    public void loop(MasterContext cxt) {

        // ArrayList<Any> packets = new ArrayList<>();  
        // 模拟发送包的数据到该节点上
        
        for (int i = 0; i < ThreadLocalRandom.current().nextLong(50, 100); i++) {
            var p = Any.pack(BacktrackingFlowOuterClass.BacktrackingFlow
                    .newBuilder()
                    .setTableId(10086)
                    .build());
                    packetsManager.addPacket(p);
            // packets.add(p);
        }

        // 必须复制. raft有一直使用该list
        var alivePeers = List.copyOf(StateFactory.getRaftNode().listAlivePeers());
        if(packetsManager.size() >= 100000) {
            log.error("告警 数据流无法正常消耗: 缓冲packets:{} 直接放弃一部分数据", packetsManager.size());
            packetsManager.discardPackets(50000);
            log.debug("master({}) execute {} packets: {}", StateFactory.getServerId(), alivePeers, packetsManager.size());
            cxt.sleep(5000);
        } else {
            // log.debug("master({}) execute {} packets: {}", StateFactory.getServerId(), alivePeers, packetsManager.size());
        }
        

        if (alivePeers == null) {
            cxt.sleep(100); // 休眠100毫秒.  
            return;
        }
 
        // 等待全部反馈后才能进入下次循环
        CountDownLatch latch = new CountDownLatch(alivePeers.size());
        // 读一致性
   

        var state = StateFactory.getStateServer().getFsm().getState();
                // log.debug("masterExecute start {} {}", status, alivePeers);
                // var state = this.<State>getValue();
                if (state == null) {
                    log.error("readIndexState获取的状态为 {}", state);
                    return;
                }
                PeerId[] peers = new PeerId[alivePeers.size()]; 
                alivePeers.toArray(peers);

                // 统计 每个节点还有多少任务容量
                var canTasks = new int[alivePeers.size()];
                for(int i = 0; i < peers.length; i++) {
                    var peer = peers[i];
                    WorkerState ws = state.getWorkers().get(peer);
                    if (ws == null) {
                        log.error("WorkerState获取的状态为 {}", ws);
                        return;
                    }
                    var can = MAX_TASKS - ws.getTaskQueueSize();
                    canTasks[i] = can;
                }
                

                // 统计每个节点发送多少任务
                var allocTasks = Utils.allocationTasks(packetsManager.size(), canTasks);
                for(int i = 0; i < peers.length; i++) {
                    var peer = peers[i];
            
                    WorkerState ws = state.getWorkers().get(peer);
                    if (ws == null) {
                        log.error("WorkerState获取的状态为 {}", ws);
                        latch.countDown(); // 所有没调用提前返回都需要 countDown
                        continue;
                    }

                    if(allocTasks[i] <= 0) {
                        latch.countDown();  // 所有没调用提前返回都需要 countDown
                        continue;
                    } 

                    ws.setUpdateAt(Instant.now());
                    ws.setTaskQueueSize(ws.getTaskQueueSize() + allocTasks[i]);
                    log.info("剩余能处理的任务数量[{}] :{}", peer, MAX_TASKS - ws.getTaskQueueSize());
                    // ws.setTaskQueueSize(MAX_TASKS);

                    var packets = packetsManager.popPackets(allocTasks[i]);
                    // 先提交 节点的 剩余能处理的任务数量. 然后再处理
                    var request = new PacketsRequest(); // 数据包切片
                    request.setPackets(packets);
                    Operate.CallOperate(
                            new Operate(OperateType.PUT_WORKERSTATE, ws),
                            new GenericClosure() {
                                @Override
                                public void run(Status status) {
                                    // log.info("PacketsRequest run {}", status);
                                    try {
                                        StateFactory.rpcClientInvokeAsync(peer.getEndpoint(), request,
                                                new InvokeCallback() {
                                                    @Override
                                                    public void complete(Object result, Throwable err) {
                                                        latch.countDown();
                                                        if (err != null) {
                                                            // TODO: 如果错误, 需要让节点恢复任务处理的状态
                                                            log.debug("{}", err);
                                                        }
                                                        // log.debug("PacketsRequest: {}", result);
                                                    }
                                                }, DEFAULT_ASYNC_TIMEOUT);
                                    } catch (InterruptedException | RemotingException e) {
                                        log.info("error send packets {}", e.toString());
                                    }
                                }
                            });

                }

         
            

       

        try {
            latch.await(DEFAULT_ASYNC_TIMEOUT, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("{}", e.toString());
        }
    }

}
